/***************************************************************************
 *   Copyright (C) 2009 by Herbert Roider   *
 *   herbert.roider@utanet.at   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *

 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef BALLISTIC_H
#define BALLISTIC_H

#include "kspline.h"
#include "utility.h"
#include <vector>

/**
 all ballistic functions

 *@author Herbert Roider
 */
namespace arw {

struct arrow {

	/** Form Factor:
	 * <pre>
	 *                         cd
	 *  f = -------------------------------------------------
	 *      cd of the Standard bullet defined in the 'G1' Model
	 *  </pre>*/
	double ff_frontal;
	/** Form Factor crosswise @see arrow#ff_frontal */
	double ff_cross;
	/** Frontal Area of the arrow [m²]*/
	double A_frontal;
	/** crosswise cutting plane of the arrow. This includes the fletches.  [m²]*/
	double A_cross;

	/** Distance from center of graphity (CG) to the center of presure (CP).
	 This is normally a negative value in meter.
	 <pre>
	 *                            cp
	 *     ____               |<------>|
	 *   __\\___\\______________|________|____________
	 *   |___________________CP________CG___________||>>>>>
	 *     /___/
	 *
	 *                              y-axis
	 *                                 |
	 *                                 |/
	 *                 ----------------|--------------> x-axis
	 *                                /|
     *                               /
     *                              /
     *                           z-axis
	 </pre>

	 For arrows this is normally a negativ value.
	 */
	double cp;
	/** Moment of inertia [kg/m²]*/
	double J;
	/** Mass of the arrow [kg] */
	double m;

};

struct flightparam {

	struct arrow arrow;

	/** air density [kg/m³] */
	double rho;

	/** Windgeschwindigkeit Geschwindigkeitsvector (zeigt in Richtung, in der er bläst)


	 */
	utility::vector_3d wind;
	/**
	 earth acceleration [m/s²]
	 Hier gibt man die Schwerkraft als Vector an. Bei einem horizontalen Schuß braucht man nur die y-Komponente. Der Vector zeigt in die Richtung, in die der Schwerkraft
	 wirkt, also bei einem horiontalen Schuss genau nach unten.

	 For a horizontal shoot only the y-axis is -9.81, and the x and z-axis is zero.


	 */
	utility::vector_3d gravity;

};

struct accel_snapshot {

	/** Translatory accelaration:
	 *
	 */
	utility::vector_3d a;
	/** angular accelaration
	 *
	 */
	utility::vector_3d alpha;
};

/**
 Flightparam, measured form the origin (Point, where the arrow leave the
 crossbow.
 */
struct flight_snapshot {
	/** distance  (m)*/
	utility::vector_3d s;

	/** velocity (m/s)*/
	utility::vector_3d v;

	/** radial velocity (rad/s):
	 */
	utility::vector_3d omega;

	/**
	 angle of the arrow [rad]
	 */
	utility::vector_3d phi;

	/**
	 No need to set this values, because some functions override this values!
	 (m/s²)
	 */
	utility::vector_3d a;

	/**
	 No need to set this values, because some functions override this values!
	 (rad/s²)
	 */
	utility::vector_3d alpha;

	/**
	 flighttime (sec)
	 */
	double t;
};

/**
 Usage:
 \code
 	// init the arrow
 	arw::arrow a;
	a.A_frontal = 0.008*0.008 * arw::utility::PI / 4.0;
	a.A_cross = 0.008* 0.5 + 0.01 * 0.01 * 5.0;
	a.cp = -0.07;
	a.ff_cross = 1.0;
	a.ff_frontal = 1.0;
	a.m = 0.027;
	a.J = 0.0001;

	arw::flightparam fp;
	arw::Ballistic::flightparam_init(fp, a);
	arw::Ballistic::flightparam_wind(fp, 90.0 * arw::utility::PI / 180.0, 5.0);// wind from side with 5m/s
	arw::Ballistic::flightparam_gravity(fp, 0.0); // horizontal shoot

    arw::Ballistic b(fp);

    // release parameter for the flight.
    // all values are measured from the global coordinate system.
    arw::flight_snapshot snapshot;
    snapshot.t = 0.0;
    snapshot.s.x = 0.0;
    snapshot.s.y = 0.0;
    snapshot.s.z = 0.0;
    snapshot.v.x = 70.0; // speed in horizontal direction
    snapshot.v.y = 0.0;
    snapshot.v.z = 0.0;
    // arrow points in flight direction:
    snapshot.phi.x = 0.0;
    snapshot.phi.y = 0.0;
    snapshot.phi.z = 0.0;
    // no rotation of the arrow:
    snapshot.omega.x = 0.0;
    snapshot.omega.y = 0.0;
    snapshot.omega.z = 0.0;

\endcode


 global system:<br \>
 the x-axis is the direcion from the crossbow to the target,<br \>
 the y-axis is vertikal direction and<br \>
 the z-axis is the cross direction, from which the wind blows.<br \>

 the origin is the origin of the crossbow. Normally the end of the rail for the arrow,
 where the arrow leaves the crossbow.


 arrow system:
 the origin is the center of gravity.
 @see arrow#cp

 http://en.wikipedia.org/wiki/Cartesian_coordinates

 @short encapsulate all ballistic functions. After init this class, you cannot change the values flichtparam or flight_const. For this case make a new object.
 */
class Ballistic {
	/**
	 *    is private, because the flightparam and m_max_stepsize has to be init.
	 * @return
	 */
	Ballistic(); //darf nicht von aussen aufgerufen werden, weil dann die flightconst nicht initiiert wird!!

	//protected:
public:
	flightparam m_flightparam;
	/**
	 initiert the cd-spline */
	static int setup_spline_cd(SplineCurve & spline);
	void get_accel(struct accel_snapshot & accel,
			const struct flight_snapshot & snapshot);
	void do_euler_step(struct flight_snapshot * snapshot, double step);
	void do_rk_step(struct flight_snapshot * snapshot, double step);

	/** only used to calc the max range */
	int calc_horizontal_angle_for_zero_range(flight_snapshot & end_fs,
			flight_snapshot & begin_fs, double v, double vertical_angle,
			double stepsize);

public:
	SplineCurve m_cd;
	Ballistic(flightparam &fp);
	//~Ballistic();
	int flightparam_at_range(flight_snapshot & sol_fs, flight_snapshot curr_fs,
			double range_x, double stepsize);

	int flightparam_at_zero_range(flight_snapshot & sol_fs,
			flight_snapshot curr_fs, double stepsize);

	int calc_max_range(flight_snapshot & begin_fs, flight_snapshot & end_fs,
			double v);

	int calc_release_parameter(flight_snapshot & solution, // release parameter
			flight_snapshot & endvalues, // impact parameter
			double range, //m
			double dV //m/s

			);

	int calc_flight_table(std::vector<flight_snapshot> &table, double max_time, // s
			double max_range, // m
			double zero_range, double muzzle_velocity);

	static void print_flight_snapshot(flight_snapshot & snapshot);
	static void print_flight_snapshot_header();

	static struct flight_snapshot linint(const struct flight_snapshot & fs1,
			const struct flight_snapshot & fs2, double fact);
	static void flightparam_gravity(flightparam & fp, double angle);
	static void flightparam_wind(flightparam &fp, double angle, double speed);
	static void flightparam_init(flightparam &fp, const arrow & a);
	/**
	 * the stepsize in seconds of one step.
	 * It should be less than 0.02 sec. because of troubles with the rotation of the arrow.
	 *
	 */
	double m_max_stepsize;

};
}
; // arw

#endif
